///////////////////////////////////////////////////////////////////////////////
// File: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2006 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "SocketServer.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/DebugTrace.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"

#include "JetByteTools/OpenSSLTools/Context.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::Output;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::ToString;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::IIndexedOpaqueUserData;

using JetByteTools::IO::IIOPool;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IAllocateBuffers;

using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateSequencedStreamSockets;
using JetByteTools::Socket::IFullAddress;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::SocketBufferSize;
using JetByteTools::Socket::ZeroByteReadConfiguration;
using JetByteTools::Socket::CConnectionLimiter;

using JetByteTools::OpenSSL::CContext;
using JetByteTools::OpenSSL::CApplicationProtocolSelection;

///////////////////////////////////////////////////////////////////////////////
// CSocketServer
///////////////////////////////////////////////////////////////////////////////

CSocketServer::CSocketServer(
   const CContext &sslContext,
   const bool verifyPeer,
   const IFullAddress &address,
   const ListenBacklog listenBacklog,
   IIOPool &pool,
   IAllocateSequencedStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   const SocketBufferSize recvBufferSize,
   const SocketBufferSize sendBufferSize,
   const ZeroByteReadConfiguration zeroByteReadConfiguration,
   const CApplicationProtocolSelection &protocolSelection)
   :  Base(
         sslContext,
         verifyPeer,
         address,
         listenBacklog,
         *this,
         pool,
         socketAllocator,
         bufferAllocator,
         recvBufferSize,
         sendBufferSize,
         zeroByteReadConfiguration,
         CConnectionLimiter::NoLimitLimiter),
      m_pProtocolSelection(&protocolSelection)
{
}

CSocketServer::CSocketServer(
   const CContext &sslContext,
   const bool verifyPeer,
   const IFullAddress &address,
   const ListenBacklog listenBacklog,
   IIOPool &pool,
   IAllocateSequencedStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   const SocketBufferSize recvBufferSize,
   const SocketBufferSize sendBufferSize,
   const ZeroByteReadConfiguration zeroByteReadConfiguration)
   :  Base(
         sslContext,
         verifyPeer,
         address,
         listenBacklog,
         *this,
         pool,
         socketAllocator,
         bufferAllocator,
         recvBufferSize,
         sendBufferSize,
         zeroByteReadConfiguration),
      m_pProtocolSelection(nullptr)
{
}

CSocketServer::~CSocketServer()
{
   Base::WaitForShutdownToComplete();
}

void CSocketServer::OnSecureConnectionEstablished(
   IStreamSocket & /*socket*/,
   const IAddress & /*address*/)
{
   DEBUG_ONLY(Output(_T("OnSecureConnectionEstablished")));
}

void CSocketServer::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   DEBUG_ONLY(Output(_T("OnConnectionEstablished")));

   socket.TryRead();
}

void CSocketServer::OnSocketReleased(
   IIndexedOpaqueUserData & /*userData*/)
{
   DEBUG_ONLY(Output(_T("OnSocketReleased")));
}

void CSocketServer::OnSecureSocketReleased(
   IIndexedOpaqueUserData & /*userData*/)
{
   DEBUG_ONLY(Output(_T("OnSecureSocketReleased")));
}

void CSocketServer::OnError(
   const _tstring &message)
{
   OutputEx(_T("OnError - ") + message);
}

void CSocketServer::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   EchoMessage(socket, buffer);

   socket.TryRead();
}

// This is where you decide if the certificate chain provided by the client
// is acceptable to identify the client.
//
// Verify is called for each certificate in the chain, starting with the least
// specific, that is, you get the root CA certificate first (as depth = x) and then
// if that's OK, you get passed other CA certs until you get the actual certificate
// that they're all vouching for (as depth = 0), something like this...
//
// Trusted Root CA, CA, CA, client cert etc.
// You can do as much or as little validation as you like here.
//
// Read the OpenSSL docs and "Network Security with OpenSSL" for details.
//
// The following is an example which is very lax!!!

int CSocketServer::Verify(
   const int preverify_ok,
   X509_STORE_CTX *pContext)
{
   auto ok = preverify_ok ? true : false;

   Output(_T("CStreamSocketConnectionManager::Verify() - do certificate verification here"));

   const auto pCert = X509_STORE_CTX_get_current_cert(pContext);
   const auto err = X509_STORE_CTX_get_error(pContext);
   const auto depth = X509_STORE_CTX_get_error_depth(pContext);

   char nameBuffer[256];

   X509_NAME_oneline(X509_get_subject_name(pCert), nameBuffer, sizeof(nameBuffer));

   Output(_T("Verify: depth = ") + JetByteTools::Core::ToString(depth) +
      _T(" err = ") + JetByteTools::Core::ToString(err) +
      _T(" - ") + CStringConverter::AtoT(X509_verify_cert_error_string(err)) +
      _T(" - ") + CStringConverter::AtoT(nameBuffer));

   const auto verify_depth = 5;      // arbitrary length for example purposes

   if (depth > verify_depth)
   {
      // set a custom error!
      X509_STORE_CTX_set_error(pContext, X509_V_ERR_CERT_CHAIN_TOO_LONG);

      ok = false; // fail the connection
   }

   if (!ok)
   {
      // OpenSSL considers the certificate at this depth to be invalid, do you agree?

      OutputEx(_T("Verify error"));

      if (err == X509_V_ERR_CERT_HAS_EXPIRED)
      {
         Output(_T("Allowing use of an expired certificate"));

         ok = true;
      }

      if (err == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT)
      {
         Output(_T("Allowing the use of self signed certificate with no issuer"));

         ok = true;
      }

      if (err == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT && pCert)
      {
         X509_NAME_oneline(X509_get_issuer_name(pCert), nameBuffer, sizeof(nameBuffer));

         Output(_T("Ignoring the fact that we can't get the issue's cert for: ") + CStringConverter::AtoT(nameBuffer));

         ok = true;
      }
   }

   if (ok)
   {
      if (depth == 0)
      {
         // The usual approach is for the subject name to match the hostname of the
         // client that's connecting - to validate that we'd need to know the name of the
         // host name of the connecting client in here. You'd grab the client address in
         // OnSecureConnectionEstablished() then do a reverse dns lookup to get the host name...

         X509_NAME_get_text_by_NID(X509_get_subject_name(pCert), NID_commonName, nameBuffer, sizeof(nameBuffer));

         //if (0 != strnicmp(nameBuffer, (LPCSTR)m_host, m_host.GetLength()))
         //{
         //   // set a custom error!
         //   X509_STORE_CTX_set_error(pContext, X509_V_ERR_CERT_REJECTED);

         //   ok = false;
         //}
      }
   }

   return ok;
}

const CApplicationProtocolSelection *CSocketServer::OnApplicationLayerProtocolNegotiation(
   IStreamSocket &socket,
   const CApplicationProtocolSelection &offeredProtocols)
{
   (void)socket;

   OutputEx(_T("ALPN protocols advertised by client: ") + offeredProtocols.GetProtocolsAsString());

   if (m_pProtocolSelection)
   {
      OutputEx(_T("ALPN protocols configured for server: ") + m_pProtocolSelection->GetProtocolsAsString());
   }
   else
   {
      OutputEx(_T("No ALPN protocols configured for server"));
   }

   // replace the default protocol selection (that was set in the constructor
   // for all connections) with a custom one for this connection, if you want...

   // return true to use the selection (either the original passed in, or the
   // new one that you've attached) or false if you don't want to support
   // ALPN on this connection

   return m_pProtocolSelection;
}

void CSocketServer::OnApplicationLayerProtocolNegotiated(
   IStreamSocket &socket,
   const _tstring &protocol)
{
   (void)socket;

   OutputEx(_T("APLN protocol selected: ") + protocol);
}

void CSocketServer::EchoMessage(
   IStreamSocket &socket,
   CSmartBuffer &buffer) const
{
   //DEBUG_ONLY(Output(_T("EchoMessage: ") + ToString(buffer->GetUsed()) + _T(" bytes")));
   //DEBUG_ONLY(Output(_T("\r\n") + DumpData(buffer->GetMemory(), buffer->GetUsed(), 60, true)));

   socket.TryWrite(buffer);
}

///////////////////////////////////////////////////////////////////////////////
// End of file: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
