///////////////////////////////////////////////////////////////////////////////
// File: MessageCollection.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2005 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "MessageCollection.h"
#include "TestMessages.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/WaitableCounter.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::CException;

///////////////////////////////////////////////////////////////////////////////
// CMessageCollection
///////////////////////////////////////////////////////////////////////////////

CMessageCollection::CMessageCollection(
   const long numberOfConnections)
   :  m_numberOfConnections(numberOfConnections),
      m_connectionsEstablished(0),
      m_connectionCounter(numberOfConnections),
      m_messageCounter(numberOfConnections)
{
   m_messages.reserve(numberOfConnections);
}

CMessageCollection::~CMessageCollection()
{
   for (auto message : m_messages)
   {
      delete message;
   }
}

void CMessageCollection::Add(
   CTestMessages *pMessages)
{
   if (!pMessages)
   {
      throw CException(_T("CMessageCollection::Add()"), _T("pMessages must not be null"));
   }

   m_messages.push_back(pMessages);
}

bool CMessageCollection::CheckMessages(
   StringErrors &errors)
{
   auto ok = true;

   for (auto message : m_messages)
   {
      if (!message->Succeeded(errors) && ok)
      {
         ok = false;
      }
   }

   return ok;
}

long CMessageCollection::ConnectionsEstablished() const
{
   return m_connectionsEstablished;
}

long CMessageCollection::ConnectionsFailed() const
{
   return m_numberOfConnections - m_connectionsEstablished;
}

long CMessageCollection::GetNumberOfConnectionsWithMessagesPending() const
{
   return m_messageCounter.GetValue();
}

void CMessageCollection::MessagesComplete()
{
   m_messageCounter.Decrement();
}

void CMessageCollection::ConnectionComplete(
   const bool succeeded)
{
   if (succeeded)
   {
      ::InterlockedIncrement(&m_connectionsEstablished);
   }

   m_connectionCounter.Decrement();
}

HANDLE CMessageCollection::GetConnectionsCompleteEvent()
{
   return m_connectionCounter.GetAtZeroWaitObject().GetWaitHandle();
}

HANDLE CMessageCollection::GetMessagesCompleteEvent()
{
   return m_messageCounter.GetAtZeroWaitObject().GetWaitHandle();
}

CMessageCollection::Iterator CMessageCollection::Begin() const
{
   return m_messages.begin();
}

CMessageCollection::Iterator CMessageCollection::End() const
{
   return m_messages.end();
}

DWORD CMessageCollection::GetAverageRTT() const
{
   DWORD averageRTT = 0;

   for (auto message : m_messages)
   {
      averageRTT += message->GetAverageRTT();
   }

   return averageRTT / static_cast<DWORD>(m_messages.size());
}

DWORD CMessageCollection::GetMaxRTT() const
{
   DWORD maxRTT = 0;

   for (auto message : m_messages)
   {
      const auto rtt = message->GetMaxRTT();

      if (rtt > maxRTT)
      {
         maxRTT = rtt;
      }
   }

   return maxRTT;
}

DWORD CMessageCollection::GetMinRTT() const
{
   DWORD minRTT = 0;

   for (auto message : m_messages)
   {
      const auto rtt = message->GetMinRTT();

      if (rtt < minRTT)
      {
         minRTT = rtt;
      }
   }

   return minRTT;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: MessageCollection.cpp
///////////////////////////////////////////////////////////////////////////////
