///////////////////////////////////////////////////////////////////////////////
// File: StreamSocketConnectionManager.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2005 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "StreamSocketConnectionManager.h"
#include "TestMessages.h"

#include "JetByteTools/IOTools/Buffer.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/WaitableCounter.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/ErrorCodeToErrorMessage.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/IAllocateSequencedStreamSockets.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::Output;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::CException;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::ToStringA;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::IQueueTimers;
using JetByteTools::Core::ErrorCodeToErrorMessage;
using JetByteTools::Core::ErrorCodeToErrorMessageA;
using JetByteTools::Core::ILockableObject;

using JetByteTools::IO::IIOPool;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IAllocateBuffers;

using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateSequencedStreamSockets;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ShutdownSend;
using JetByteTools::Socket::ConnectionClosureReason;
using JetByteTools::Socket::ConnectionAbortedMask;
using JetByteTools::Socket::GetClosureReasonAsString;

using JetByteTools::OpenSSL::CContext;
using JetByteTools::OpenSSL::CApplicationProtocolSelection;

using std::cout;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// CStreamSocketConnectionManager
///////////////////////////////////////////////////////////////////////////////

CStreamSocketConnectionManager::CStreamSocketConnectionManager(
   const CContext &sslContext,
   const bool verifyPeer,
   const Milliseconds readTimeout,
   IIOPool &pool,
   IAllocateSequencedStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   const CApplicationProtocolSelection &protocolSelection)
   :  JetByteTools::OpenSSL::CStreamSocketConnectionManager(sslContext, verifyPeer, *this, pool, socketAllocator, bufferAllocator),
      m_timerHandleIndex(socketAllocator.RequestUserDataSlot(_T("CStreamSocketConnectionManager - m_timerHandleIndex"))),
      m_userDataIndex(socketAllocator.RequestUserDataSlot(_T("CStreamSocketConnectionManager - m_userDataIndex"))),
      m_timeoutFilter(*this, m_timerQueue),
      m_readTimeout(readTimeout),
      m_pProtocolSelection(&protocolSelection)
{

}

CStreamSocketConnectionManager::CStreamSocketConnectionManager(
   const CContext &sslContext,
   const bool verifyPeer,
   const Milliseconds readTimeout,
   IIOPool &pool,
   IAllocateSequencedStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator)
   :  JetByteTools::OpenSSL::CStreamSocketConnectionManager(sslContext, verifyPeer, *this, pool, socketAllocator, bufferAllocator),
      m_timerHandleIndex(socketAllocator.RequestUserDataSlot(_T("CStreamSocketConnectionManager - m_timerHandleIndex"))),
      m_userDataIndex(socketAllocator.RequestUserDataSlot(_T("CStreamSocketConnectionManager - m_userDataIndex"))),
      m_timeoutFilter(*this, m_timerQueue),
      m_readTimeout(readTimeout),
      m_pProtocolSelection(nullptr)
{

}

CStreamSocketConnectionManager::~CStreamSocketConnectionManager()
{
   WaitForShutdownToComplete();
}

void CStreamSocketConnectionManager::OnPreOutgoingConnect(
   IIndexedOpaqueUserData &userData,
   const IAddress & /*address*/,
   const void *pUserData)
{
   const auto pTestMessages = const_cast<CTestMessages *>(static_cast<const CTestMessages*>(pUserData));

   userData.SetUserPointer(m_userDataIndex, pTestMessages);
}

void CStreamSocketConnectionManager::OnSecureConnectionEstablished(
   IStreamSocket & /*socket*/,
   const IAddress & /*address*/)
{
   DEBUG_ONLY(Output(_T("OnSecureConnectionEstablished")));
}

void CStreamSocketConnectionManager::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   DEBUG_ONLY(Output(_T("OnConnectionEstablished")));

   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   pTestMessages->ConnectionComplete(true);

   const auto timerHandle = m_timerQueue.CreateTimer();

   socket.SetUserData(m_timerHandleIndex, timerHandle);

   if (m_readTimeout != 0)
   {
      m_timeoutFilter.SetReadTimeout(socket, m_readTimeout, *this, 1);
   }

   socket.TryRead();

   SendData(socket);
}

void CStreamSocketConnectionManager::OnOutgoingConnectionFailed(
   IStreamSocket &socket,
   const IAddress & /*address*/,
   const DWORD lastError)
{
   {
      ILockableObject::Owner lock(m_lock);

      m_connectionErrors[lastError]++;
   }

   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   socket.SetUserData(m_timerHandleIndex, IQueueTimers::InvalidHandleValue);

   pTestMessages->ConnectionComplete(false);

   socket.SetUserPointer(m_userDataIndex, nullptr);
}

void CStreamSocketConnectionManager::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   DEBUG_ONLY(Output(_T("OnSocketReleased")));

   const auto timerHandle = userData.GetUserData(m_timerHandleIndex);

   if (timerHandle != IQueueTimers::InvalidHandleValue)
   {
      if (m_timerQueue.DestroyTimer(timerHandle))
      {
         OnError(_T("CStreamSocketConnectionManager::OnSocketReleased() - timer was set when destroyed, socket ref count bug!"));
      }
   }
}

void CStreamSocketConnectionManager::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   ProcessData(socket, buffer);

   socket.TryRead(buffer);
}

void CStreamSocketConnectionManager::OnConnectionReset(
   IStreamSocket & /*socket*/,
   const DWORD lastError)
{
   ILockableObject::Owner lock(m_lock);

   m_connectionResetErrors[lastError]++;
}

void CStreamSocketConnectionManager::OnError(
   const _tstring &message)
{
   ILockableObject::Owner lock(m_lock);

   m_errors[message]++;
}

void CStreamSocketConnectionManager::OnError(
   const _tstring &message,
   const DWORD lastError)
{
   OnError(message + _T(": ") + ErrorCodeToErrorMessage(lastError));
}

void CStreamSocketConnectionManager::OnConnectionClosure(
   IStreamSocket &socket,
   const ConnectionClosureReason reason)
{
   const auto timerHandle = socket.GetUserData(m_timerHandleIndex);

   socket.SetUserData(m_timerHandleIndex, IQueueTimers::InvalidHandleValue);

   if (timerHandle != IQueueTimers::InvalidHandleValue)
   {
      m_timerQueue.DestroyTimerWithRefCountedUserData(timerHandle, &socket);
   }

   if (reason & ConnectionAbortedMask)
   {
      OnError(GetClosureReasonAsString(reason));
   }

   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   if (pTestMessages)
   {
      pTestMessages->MessagesComplete(!(reason & ConnectionAbortedMask));
   }
}

void CStreamSocketConnectionManager::ProcessData(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   if (pTestMessages)
   {
      try
      {
         auto messageComplete = false;

         if (pTestMessages->ValidateReceivedData(buffer->GetMemory(), buffer->GetUsed(), messageComplete))
         {
            if (messageComplete && pTestMessages->SendAfterRecv())
            {
               SetSendTimer(socket, pTestMessages->GetTransferRateMillis());
            }
         }
         else if (messageComplete)
         {
            const auto holdConnections = pTestMessages->HoldConnections();

            if (!holdConnections)
            {
               socket.Shutdown(ShutdownSend);
            }
            else
            {
               pTestMessages->MessagesComplete(true);
            }
         }
      }
      catch (CException &e)
      {
         ILockableObject::Owner lock(m_lock);

         m_errors[e.GetWhat()]++;

         pTestMessages->MessagesComplete(false);

         throw;
      }
   }

   buffer->Empty();
}

void CStreamSocketConnectionManager::OnTimer(
   IQueueTimers::Timer::UserData userData)
{
   auto &socket = *reinterpret_cast<IStreamSocket *>(userData);

   SendData(socket);

   socket.Release();
}

void CStreamSocketConnectionManager::OnTimer(
   IStreamSocket &socket,
   TimerCallback::UserData /*userData*/)
{
   OnError(_T("Read timeout"));

   socket.AbortConnection();
}

void CStreamSocketConnectionManager::SendData(
   IStreamSocket &socket)
{
   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   if (pTestMessages)
   {
      if (pTestMessages->WriteSome(socket))
      {
         if (!pTestMessages->SendAfterRecv())
         {
            SetSendTimer(socket, pTestMessages->GetTransferRateMillis());
         }
      }
      else if (!pTestMessages->HoldConnections())
      {
         socket.Shutdown(ShutdownSend);
      }
   }
}

void CStreamSocketConnectionManager::SetSendTimer(
   IStreamSocket &socket,
   const Milliseconds timeout)
{
   const auto timerHandle = socket.GetUserData(m_timerHandleIndex);

   m_timerQueue.SetTimerWithRefCountedUserData(timerHandle, *this, timeout, &socket);
}

bool CStreamSocketConnectionManager::DumpErrors() const
{
   auto ok = true;

   if (!DumpErrors(_T("Connection establishment errors"), m_connectionErrors))
   {
      ok = false;
   }

   if (!DumpErrors(_T("Connections reset"), m_connectionResetErrors))
   {
      ok = false;
   }

   if (!m_errors.empty())
   {
      ok = false;

      cout << "Other errors" << endl;

      for (const auto &error : m_errors)
      {
         cout << ToStringA(error.second) << " - " << CStringConverter::TtoA(error.first) << endl;
      }
   }

   return ok;
}

bool CStreamSocketConnectionManager::DumpErrors(
   const _tstring &errorType,
   const LastErrors &lastErrors) const
{
   auto ok = true;

   if (!lastErrors.empty())
   {
      ok = false;

      cout << CStringConverter::TtoA(errorType) << endl;

      for (const auto &lastError : lastErrors)
      {
         cout << ToStringA(lastError.second) << " - " << ErrorCodeToErrorMessageA(lastError.first) << endl;
      }
   }

   return ok;
}

// This is where you decide if the certificate chain provided by the client
// is acceptable to identify the client.
//
// Verify is called for each certificate in the chain, starting with the least
// specific, that is, you get the root CA certificate first (as depth = x) and then
// if that's OK, you get passed other CA certs until you get the actual certificate
// that they're all vouching for (as depth = 0), something like this...
//
// Trusted Root CA, CA, CA, client cert etc.
// You can do as much or as little validation as you like here.
//
// Read the OpenSSL docs and "Network Security with OpenSSL" for details.
//
// The following is an example which is very lax!!!

int CStreamSocketConnectionManager::Verify(
   const int preverify_ok,
   X509_STORE_CTX *pContext)
{
   auto ok = preverify_ok ? true : false;

   Output(_T("CStreamSocketConnectionManager::Verify() - do certificate verification here"));

   const auto pCert = X509_STORE_CTX_get_current_cert(pContext);
   const auto err = X509_STORE_CTX_get_error(pContext);
   const auto depth = X509_STORE_CTX_get_error_depth(pContext);

   char nameBuffer[256];

   X509_NAME_oneline(X509_get_subject_name(pCert), nameBuffer, sizeof(nameBuffer));

   Output(_T("Verify: depth = ") + JetByteTools::Core::ToString(depth) +
      _T(" err = ") + JetByteTools::Core::ToString(err) +
      _T(" - ") + CStringConverter::AtoT(X509_verify_cert_error_string(err)) +
      _T(" - ") + CStringConverter::AtoT(nameBuffer));

   const auto verify_depth = 5;      // arbitrary length for example purposes

   if (depth > verify_depth)
   {
      // set a custom error!
      X509_STORE_CTX_set_error(pContext, X509_V_ERR_CERT_CHAIN_TOO_LONG);

      ok = false; // fail the connection
   }

   if (!ok)
   {
      // OpenSSL considers the certificate at this depth to be invalid, do you agree?

      Output(_T("Verify error"));

      if (err == X509_V_ERR_CERT_HAS_EXPIRED)
      {
         Output(_T("Allowing use of an expired certificate"));

         ok = true;
      }

      if (err == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT)
      {
         Output(_T("Allowing the use of self signed certificate with no issuer"));

         ok = true;
      }

      if (err == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT && pCert)
      {
         X509_NAME_oneline(X509_get_issuer_name(pCert), nameBuffer, sizeof(nameBuffer));

         Output(_T("Ignoring the fact that we can't get the issue's cert for: ") + CStringConverter::AtoT(nameBuffer));

         ok = true;
      }
   }

   if (ok)
   {
      if (depth == 0)
      {
         // The usual approach is for the subject name to match the hostname of the
         // server that you are connecting to- to validate that we'd need to know the name of the
         // host name of the connecting server in here....

         X509_NAME_get_text_by_NID(X509_get_subject_name(pCert), NID_commonName, nameBuffer, sizeof(nameBuffer));

         //if (0 != strnicmp(nameBuffer, (LPCSTR)m_host, m_host.GetLength()))
         //{
         //   // set a custom error!
         //   X509_STORE_CTX_set_error(pContext, X509_V_ERR_CERT_REJECTED);

         //   ok = false;
         //}
      }
   }

   return ok;
}

const CApplicationProtocolSelection *CStreamSocketConnectionManager::GetApplicationProtocolSelection(
   IStreamSocket &socket)
{
   (void)socket;

   if (m_pProtocolSelection)
   {
      OutputEx(_T("ALPN protocols advertised: ") + m_pProtocolSelection->GetProtocolsAsString());
   }

   return m_pProtocolSelection;
}

const CApplicationProtocolSelection *CStreamSocketConnectionManager::OnApplicationLayerProtocolNegotiation(
   IStreamSocket &socket,
   const CApplicationProtocolSelection &offeredProtocols)
{
   (void)socket;

   OutputEx(_T("ALPN protocols advertised by client: ") + offeredProtocols.GetProtocolsAsString());

   if (m_pProtocolSelection)
   {
      OutputEx(_T("ALPN protocols configured for server: ") + m_pProtocolSelection->GetProtocolsAsString());
   }
   else
   {
      OutputEx(_T("No ALPN protocols configured for server"));
   }

   return m_pProtocolSelection;
}

void CStreamSocketConnectionManager::OnApplicationLayerProtocolNegotiated(
   IStreamSocket &socket,
   const _tstring &protocol)
{
   (void)socket;

   OutputEx(_T("APLN protocol selected: ") + protocol);
}

///////////////////////////////////////////////////////////////////////////////
// End of file: StreamSocketConnectionManager.cpp
///////////////////////////////////////////////////////////////////////////////
