///////////////////////////////////////////////////////////////////////////////
// File: RoundTripTimes.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2010 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "RoundTripTimes.h"

#include "JetByteTools/CoreTools/Exception.h"

#pragma hdrstop

#include <limits>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::CException;

///////////////////////////////////////////////////////////////////////////////
// CRoundTripTimes
///////////////////////////////////////////////////////////////////////////////

CRoundTripTimes::CRoundTripTimes()
{
   SetNumberOfTimes(1);
}

void CRoundTripTimes::SetNumberOfTimes(
   const size_t numberOfTimes)
{
   if (numberOfTimes == 0)
   {
      throw CException(_T("CRoundTripTimes::SetNumberOfTimes()"), _T("Unexpected; numberOfTimes cannot be zero"));
   }

   m_times.resize(numberOfTimes);
}

void CRoundTripTimes::MarkAsSent(
   const size_t index)
{
   if (index >= m_times.size())
   {
      throw CException(_T("CRoundTripTimes::MarkAsSent()"), _T("index out of range"));
   }

   m_times[index].MarkAsSent();
}

void CRoundTripTimes::MarkAsReceived(
   const size_t index)
{
   if (index >= m_times.size())
   {
      throw CException(_T("CRoundTripTimes::MarkAsSent()"), _T("index out of range"));
   }

   m_times[index].MarkAsReceived();
}

CRoundTripTimes::Iterator CRoundTripTimes::Begin() const
{
   return m_times.begin();
}

CRoundTripTimes::Iterator CRoundTripTimes::End() const
{
   return m_times.end();
}

size_t CRoundTripTimes::GetNumSent() const
{
   size_t count = 0;

   for (Times::const_iterator it = m_times.begin(), end = m_times.end();
      it != end;
      ++it)
   {
      if (it->GetStatus() & CRoundTripTime::Sent)
      {
         count++;
      }
   }

   return count;
}

size_t CRoundTripTimes::GetNumReceived() const
{
   size_t count = 0;

   for (Times::const_iterator it = m_times.begin(), end = m_times.end();
      it != end;
      ++it)
   {
      if (it->GetStatus() == CRoundTripTime::Received)
      {
         count++;
      }
   }

   return count;
}

DWORD CRoundTripTimes::GetAverageRTT() const
{
   size_t average = 0;
   size_t count = 0;

   for (Times::const_iterator it = m_times.begin(), end = m_times.end();
      it != end;
      ++it)
   {
      if (it->GetStatus() == CRoundTripTime::Received)
      {
         count++;

         average += it->GetRoundTripTime();
      }
   }

   return count != 0 ? static_cast<DWORD>(average / count) : 0;
}

DWORD CRoundTripTimes::GetMaxRTT() const
{
   DWORD max = 0;

   for (Times::const_iterator it = m_times.begin(), end = m_times.end();
      it != end;
      ++it)
   {
      if (it->GetStatus() == CRoundTripTime::Received)
      {
         const DWORD rtt = it->GetRoundTripTime();

         if (max < rtt)
         {
            max = rtt;
         }
      }
   }

   return max;
}

DWORD CRoundTripTimes::GetMinRTT() const
{
   DWORD min = std::numeric_limits<DWORD>::max();

   for (Times::const_iterator it = m_times.begin(), end = m_times.end();
      it != end;
      ++it)
   {
      if (it->GetStatus() == CRoundTripTime::Received)
      {
         const DWORD rtt = it->GetRoundTripTime();

         if (min > rtt)
         {
            min = rtt;
         }
      }
   }

   return min == std::numeric_limits<DWORD>::max() ? 0 : min;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: RoundTripTimes.cpp
///////////////////////////////////////////////////////////////////////////////
